(** Barême (/60)
Questions :
- Échauffement (5 pts)
- Part A (20 pts)
- Part B (20 pts)
Rédaction (10 pts)
Tactiques (5 points)
**)
(** Pour obtenir tous les points de rédaction :
- Indentez vos codes avec des bullets points pour les disjonctions de cas.
- Nommez explicitement vos variables et hypothèses avec des noms parlants.
- Quand vous introduisez des lemmes intermédiaires, donnez leur un nom parlant 
  et expliquez succintement ce qu'ils font. Dites où vous les utilisez.
- Dans les preuves longues, rajoutez des commentaires pour rendre votre preuve 
  lisible. Par exemple, dans une disjonction de cas, avant chaque bullet point 
  (*If x = ... *).
- Essayez de trouver toujours la preuve la plus concise possible
  (si une preuve est trop longue, ou une induction trop moche, vous êtes 
  probablement sur la mauvaise voie).
  N'hésitez pas à introduire des lemmes intermédiaires
*)
(** Pour obtenir tous les points de tactiques :
- Soyez paresseux ! Si vous avez l'impression de toujours répéter les mêmes 
  commandes les unes après les autres,
  c'est l'occasion d'automatiser ça avec une tactique.
- Pas besoin de beaucoup de tactiques, ni de tactiques compliqués, pensez simple
  et utile.
- Pour chaque tactique, dites succintement ce qu'elle fait et donnez au moins un
  lemme où vous l'utilisez.
*)

Require Import List.

(** * Grammars

    Given two finite sets of terminal and non-terminal symbols,
    a grammar is a finite set of productions of the form
    [S ~> w] where [S] is a non-terminal symbol and [w] is a list of
    (terminal or non-terminal) symbols.

    We shall use lowercase identifiers for terminals (e.g. [a], [b], [x])
    and uppercase ones for non-terminals (e.g. [S], [E], [U]).

    We shall identify both kinds of symbols by natural numbers in Coq
    (in practice, only finitely many natural number identifiers will be
    used in a grammar) as shown in the next definition. *)

Definition tsym := nat.
Definition nsym := nat.

Inductive sym := T : tsym -> sym | NT : nsym -> sym.

Definition grammar : Set := list (nsym * list sym).

(** * Derivations

    Grammars define languages, through the notion of derivation:
    starting from an initial list of symbols [u], one can repeatedly expand
    occurrences of any non-terminal S into some w such that the grammar
    contains the production [S ~> w].
    The terminal words obtained in this way form the language of [u].

    One does not loose anything by restricting to left derivations,
    i.e. derivations where only leftmost non-terminals are expanded.
    It will be more convenient for us to work with this notion, which
    is defined next. *)

Inductive deriv (g:grammar) : list sym -> list sym -> Prop :=
  | deriv_nil : deriv g nil nil
  | deriv_refl : forall l w s, deriv g l w -> deriv g (s::l) (s::w)
  | deriv_NT :
      forall exp l w u,
      In (u,exp) g ->
      deriv g (exp ++ l) w ->
      deriv g (NT u :: l) w.

(** * Parsing and LL(1) grammars

    Given a terminal word, we want to efficiently determine whether the word
    belongs to the language of the initial non-terminal [NT 0]. There is an 
    obvious non-deterministic algorithm doing this, but a linear-time 
    deterministic algorithm is desirable.

    We will work on the class of LL(1) grammars, where such parsers exists, 
    and moreover the parsing function only needs to look at the next terminal 
    to determine which productions to use.

    The example grammar [ABAB.g] is not [LL(1)] because we need to look at the 
    next two symbols to know where we should use [S ~> A] or [S ~> B].
    However, [Dyck.g] is.

    We define next some notions that are useful to determine when a grammar is 
    LL(1), and build parsers for them. *)

Section LL.

  (** All definitions in this section will be parameterized by [G]. *)
  Variable G : grammar.

  (** A list of symbols is nullable if it can derive the empty word. *)
  Definition Nullable e := deriv G e nil.

  (** [First a e] means that a word in the language of [e]
      starts with the terminal [a]. *)
  Definition First a e := exists v, deriv G e (T a :: v).

  (** A list of symbols is reachable if it is the suffix of a list
      that can be derived from the initial non-terminal. *)
  Definition Reachable u := exists w, deriv G (NT 0 :: nil) (w ++ u).

  Definition Follow a U := exists v, Reachable (NT U :: v) /\ First a v.

  Definition End U := exists v, Reachable (NT U :: v) /\ Nullable v.

  (** The high-level idea behind LL(1) grammars is that, when attempting to
      determine whether [w] belongs to the language of [NT U :: l], we want
      to be able to determine a unique production [U ~> e] which can make
      this work. When two productions could work, we say that there is a
      conflict.

      Conflicts can happen in several cases, for instance when [w = a::_]
      and there are productions [U ~> e] and [U ~> e'] such that [First a e]
      and [First a e']. It also happens when [First a e], [Nullable e'] and
      [Follow a U]: intuitively, we could obtain [a] from [e], but we could
      also get rid of [e'] and then obtain [a] from [l] because it may follow
      [U].

      A grammar is strongly LL(1) when there is no distinct productions
      [U ~> e] and [U ~> e'] that are in conflict, in the following sense:
      instead of definition a conjunction of conditions, we use a record,
      so that conjuncts get names. When [r : no_conflict u e e'],
      [nullable_nullable r : Nullable e -> Nullable e' -> e = e']. *)

  Record no_conflict u e e' : Prop := {
    first_first : forall a, First a e -> First a e' -> e = e' ;
    first_follow : forall a, First a e -> Nullable e' -> Follow a u -> e = e' ;
    nullable_nullable : Nullable e -> Nullable e' -> e = e'
  }.

  Definition LL_1 :=
    forall u e e', In (u,e) G -> In (u,e') G -> no_conflict u e e'.

End LL.

(** Tactic for proving membership in an explicit list. *)
Ltac auto_In := repeat (try apply in_eq ; apply in_cons).

(** Feel free to add more tactics of your own here *)

(** * Examples to get used to the notion of grammar,
    derivation, and LL(1) conflicts. *)

Module ABAB.

  (** A simple example grammar, where:
      - [S -> A]
      - [S -> B]
      - [A -> a a S]
      - [B -> a b S]
      - [B -> a b]

      We encode [S], [A] and [B] as [0], [1] and [2] respectively. *)

  Definition a := T 0.
  Definition b := T 1.

  Definition g : grammar :=
    (0, NT 1 :: nil) ::
    (0, NT 2 :: nil) ::
    (1, a :: a :: NT 0 :: nil) ::
    (2, a :: b :: NT 0 :: nil) ::
    (2, a :: b :: nil) ::
    nil.

  Goal deriv g (NT 0 :: nil) (a :: a :: a :: b :: nil).
  Proof.
    apply deriv_NT with (exp := NT 1 :: nil); [auto_In|].
    apply deriv_NT with (exp := a :: a :: NT 0 :: nil); [auto_In|].
    repeat apply deriv_refl.
    apply deriv_NT with (exp := NT 2 :: nil); [auto_In|].
    apply deriv_NT with (exp := a :: b :: nil); [auto_In|].
    repeat apply deriv_refl.
    apply deriv_nil.
  Qed.

End ABAB.

Module Dyck.

  (** We now captue the language of well-parenthesized words over an
      alphabet featuring parentheses and brackets.

      Terminals: Left/Right PARentheses/BRAckets. *)

  Definition LPAR := T 0.
  Definition RPAR := T 1.
  Definition LBRA := T 2.
  Definition RBRA := T 3.

  (** The grammar features a single non-terminal. *)

  Definition S := 0.

  Definition g : grammar :=
    (S, nil) ::
    (S, LPAR :: NT S :: RPAR :: NT S :: nil) ::
    (S, LBRA :: NT S :: RBRA :: NT S :: nil) ::
    nil.

(** À vous de jouer (5 pts) **)

(*Q1*)
  Goal deriv g (NT 0 :: nil) 
  (LPAR :: LBRA :: RBRA :: LBRA :: RBRA :: RPAR :: nil).
  Proof.
  Admitted.

  (** The dyck grammar is LL(1): we verify two representative cases. *)
(*Q2*)
  Lemma not_First_nil : forall g a, ~ First g a nil.
  Proof.
  Admitted.
(*Q3*)
  Lemma not_Nullable_T : forall g a e, ~ Nullable g (T a :: e).
  Proof.
  Admitted.
(*Q4*)
  Lemma First_T : forall g a a' e, First g a (T a' :: e) -> a = a'.
  Proof.
  Admitted.

  Goal no_conflict g 0 nil (LPAR :: NT 0 :: RPAR :: NT 0 :: nil).
  Proof.
    split; intros.
    + elim (not_First_nil _ _ H).
    + elim (not_First_nil _ _ H).
    + elim (not_Nullable_T _ _ _ H0).
  Qed.
(*Q5*)
  Goal no_conflict g 0 (LPAR :: NT 0 :: RPAR :: NT 0 :: nil)
                       (LBRA :: NT 0 :: RBRA :: NT 0 :: nil).
  Proof.
  Admitted.

End Dyck.

(** * Part A - LL(1) parsing (20 pts) **)

   (* Assume some arbitrary LL(1) grammar [G]. *)

Variable G : grammar.

Axiom G_LL1 : LL_1 G.

(** Your goal is to define and prove a parser for LL(1) grammars,
    assuming that [Nullable], [First], etc. are computable (which is
    always the case, but we won't prove it).
    More precisely, assume some functions that give us productions
    satisfying some properties based on [Nullable], [First], etc. *)

Definition option A (P:A->Prop) := { e | P e } + { forall e, ~ P e }.
(** A value of type [option A P] is:
    - either a witness [e:A] together with a proof of [P e],
    - or a proof that no such [e] exists.

    Technically we make use of [sig] and [sumor] which is similar but
    different from [sumbool]. *)

Variable find_first : forall a U,
  option _ (fun e => In (U,e) G /\ First G a e).

Variable find_follow : forall a U,
  option _ (fun e => In (U,e) G /\ Follow G a U /\ Nullable G e).

Variable find_end : forall U,
  option _ (fun e => In (U,e) G /\ End G U /\ Nullable G e).

(** ** Parsing function

    Define a function [parse : list tsym -> list sym -> nat -> bool]
    such that, for all words [w] in the language of [s],
    [parse w s n = true] for large enough values of [n].
    Of course [parse w s n = true] should only hold for words [w]
    in the language of [s].
 
    In Coq's jargon, the extra parameter [n] is the "fuel" of the
    [parse] function. It should decrease at every recursive call,
    making a [Fixpoint] definition valid, which would be very difficult
    to obtain otherwise.
     
    Parsing should proceed based on the first element of each list:
    - [parse (a::w) (a::s) f] should consume the terminals [a],
      and continue with [parse w s (f-1)];
    - [parse (a::w) (U::s) f] should use some reduction [U ~> e]
      such that [First a e], or such that [Nullable e] and [Follow a U],
      and continue with [parse (a::w) (exp++s) (f-1)];
    - [parse nil (U::_) f] should use some reduction [U ~> e]
      when [Nullable e] and [End U];
    - [parse nil nil f] should return [true];
    - other cases should return [false]. *)

Require Import Arith.
(*Q6*)
(** Implementing parse **)
Fixpoint parse (w : list tsym) (s : list sym) fuel : bool := match fuel with
  | O => false
  | S fuel =>
    match w,s with
    | _, _ => false (* TODO *)
    end
end.

(** Correctness of parse **)

    (*This part should not be too hard, as the parsing function performs
    steps allowed by [deriv]. *)

(*Q7*)
Lemma correct : forall fuel w s,
  parse w s fuel = true -> deriv G s (map T w).
Proof.
Admitted.

(** Completeness of parse **)

    (* This is significantly more difficult: given a derivation of [w],
    we need to show that [parse] necessarily follows it, notably
    selecting the same productions. This relies on the LL(1) property
    of [G]. You will need several intermediate lemmas; a few (probably)
    useful ones are given below for guidance. *)

Lemma Reachable_cons : forall hd tl, Reachable G (hd :: tl) -> Reachable G tl.
Proof.
  intros hd tl (w,H).
  exists (w ++ hd :: nil).
  (* To find standard library lemmas such as [app_assoc],
     use [SearchAbout app] or [Search (_ ++ _ ++ _)]. *)
  rewrite<- app_assoc.
  exact H.
Qed.

(*Q8*)
Lemma deriv_refl_concat : forall l w e, deriv G l w -> deriv G (e++l) (e++w).
Admitted.

(*Q9*)
(** We essentially want to show that, if [U ~> exp],
   [deriv (w ++ NT U :: l) s] implies [deriv (w ++ exp ++ l)].
   The lemma is formulated in an indirect way so that we do not loose
   information about [o] when using [induction]. *)
Lemma deriv_deep : forall U exp i o, In (U,exp) G ->
  deriv G i o -> forall w l, o = (w ++ NT U :: l) -> deriv G i (w ++ exp ++ l).
Proof.
Admitted.

Lemma Reachable_NT :
  forall U exp l, In (U,exp) G -> Reachable G (NT U :: l) ->
  Reachable G (exp ++ l).
Proof.
  intros U exp l Hexp (w,HR).
  exists w.
  apply deriv_deep with (U:=U) (exp:=exp) (o := w ++ NT U :: l); auto.
Qed.

(*Q10*)
Lemma Nullable_app : forall u v, Nullable G (u ++ v) ->
  Nullable G u /\ Nullable G v.
Proof.
Admitted.

(*Q11*)
Lemma deriv_app : forall u v t w,
  deriv G (u ++ v) (T t :: w) ->
  (Nullable G u /\ First G t v) \/ First G t u.
Proof.
Admitted.

(*Q12*)
Theorem complete :
  forall w, deriv G (NT 0 :: nil) (map T w) ->
  exists f, parse w (NT 0 :: nil) f = true.
Proof.
Admitted.

(** * Part B - Computing [Nullable] (20 pts)

    The goal now is to define a function that computes the [Nullable] predicate. *)

Require Import List.
Require Import Arith.

(** ** Inductive definition of nullable symbols

    The first step is to show that [Nullable] can be characterized inductively
    as follows. *)

Inductive Nullable_sym (g:grammar) : sym -> Prop :=
  | Nullable_In : forall nt exp, In (nt,exp) g ->
      (forall x, In x exp -> Nullable_sym g x) ->
      Nullable_sym g (NT nt).

(*Q13*)
Theorem Nullable_sym_spec : forall g s, Nullable_sym g s <-> Nullable g (s::nil).
Proof.
Admitted.

(** ** Computation step

    We first define a function [nullable_step] that takes a set of non-terminals
    assumed to be nullable, and attempts to enrich it with new symbols based
    on the productions of the grammar. *)

Require Import MSets.
Module NatSet := MSetList.Make(Nat).

(** [NatSet] allows to work on sets of natural numbers (represented as strictly
    increasing lists). Coq's standard equality [eq] won't hold between sets 
    containing the same elements but built in different ways. Set equality is 
    available as [NatSet.equal], but you should be able to complete this file 
    without using it. *)

(** Now define [all_NT_In_dec e l] which determines if all items in [e] are
    non-terminals that belong to [l]. *)
(*Q14*)
Lemma all_NT_In_dec : forall (e:list sym) (l:NatSet.t),
  { forall x, In x e -> exists y, x = NT y /\ NatSet.In y l } +
  { ~ forall x, In x e -> exists y, x = NT y /\ NatSet.In y l }.
Proof.
Admitted.

Fixpoint nullable_step (g:grammar) (s:NatSet.t) : NatSet.t :=
  match g with
  | nil => s
  | (n,e)::g =>
      if all_NT_In_dec e s then NatSet.add n (nullable_step g s) else nullable_step g s
  end.

(** You will need to prove (at least) the correctness and completeness of
    [nullable_step]. Formulating these two lemmas requires some care; the most
    immediate formulation won't allow a direct proof by induction. *)


(** ** Iterated computation *)

(** Iterate [nullable_step] some number of times, starting with [NatSet.empty].
    A possible bonus: improve the function to stop when a fixed point is 
    reached. *)
Fixpoint nullables g k : NatSet.t :=
  match k with
   | O => NatSet.empty
   | S k => nullable_step g (nullables g k)
  end.

Definition nullable_sym g s f :=
  match s with
    | T _ => False
    | NT n => NatSet.In n (nullables g f)
  end.

(** Correctness of [nullable_sym]. *)
(*Q15*)
Lemma nullable_Nullable : forall g f s, nullable_sym g s f -> Nullable_sym g s.
Proof.
Admitted.

(*Q16*)
(** Completeness of [nullable_sym]. *)
Lemma Nullable_nullable :
  forall g s, Nullable_sym g s ->
  exists f, nullable_sym g s f.
Proof.
Admitted.





